from __future__ import annotations

from abstractruntime.core.models import EffectType, RunState, RunStatus
from abstractruntime.visualflow_compiler import compile_visualflow


def _plan_for_node(node_type: str, effect_config: dict, input_data: dict | None = None) -> object:
    spec = compile_visualflow(
        {
            "id": f"vf_{node_type}",
            "name": f"vf_{node_type}",
            "entryNode": "node",
            "nodes": [
                {
                    "id": "node",
                    "type": node_type,
                    "data": {"nodeType": node_type, "effectConfig": effect_config},
                }
            ],
            "edges": [],
        }
    )
    run = RunState(
        run_id="run",
        workflow_id=str(spec.workflow_id),
        status=RunStatus.RUNNING,
        current_node="node",
        vars={"_temp": {}},
    )
    if isinstance(input_data, dict):
        run.vars["_last_output"] = dict(input_data)
    return spec.nodes["node"](run, input_data)


def test_generate_image_node_compiles_to_llm_call_output_selector() -> None:
    plan = _plan_for_node(
        "generate_image",
        {
            "prompt": "a horse",
            "runtime_provider": "openai",
            "runtime_model": "gpt-4.1-mini",
            "image_provider": "abstractvision",
            "image_model": "flux-test",
            "format": "png",
            "width": 512,
            "height": 512,
            "steps": 8,
        },
    )

    assert plan.effect is not None
    assert plan.effect.type == EffectType.LLM_CALL
    payload = dict(plan.effect.payload or {})
    output = dict(payload.get("output") or {})
    assert payload["prompt"] == "a horse"
    assert "provider" not in payload
    assert "model" not in payload
    assert output["modality"] == "image"
    assert output["task"] == "image_generation"
    assert output["format"] == "png"
    assert output["provider"] == "abstractvision"
    assert output["model"] == "flux-test"
    assert output["width"] == 512
    assert output["height"] == 512
    assert output["steps"] == 8


def test_generate_image_legacy_provider_model_are_not_media_fallbacks() -> None:
    plan = _plan_for_node(
        "generate_image",
        {
            "prompt": "a horse",
            "provider": "abstractvision",
            "model": "flux-test",
        },
    )

    assert plan.effect is not None
    payload = dict(plan.effect.payload or {})
    output = dict(payload.get("output") or {})
    assert "provider" not in payload
    assert "model" not in payload
    assert "provider" not in output
    assert "model" not in output


def test_generate_image_does_not_treat_runtime_provider_input_as_image_provider() -> None:
    plan = _plan_for_node(
        "generate_image",
        {"prompt": "a horse", "image_model": "gpt-image-1-mini"},
        {"provider": "lmstudio", "model": "chat-model"},
    )

    assert plan.effect is not None
    payload = dict(plan.effect.payload or {})
    output = dict(payload.get("output") or {})
    assert "provider" not in output
    assert output["model"] == "gpt-image-1-mini"


def test_edit_image_node_compiles_to_llm_call_image_edit_selector() -> None:
    plan = _plan_for_node(
        "edit_image",
        {
            "prompt": "Make the jacket red.",
            "image_artifact": {"$artifact": "source-img", "content_type": "image/png"},
            "mask_artifact": "mask-img",
            "image_provider": "openai-compatible",
            "image_model": "gpt-image-1",
            "format": "png",
            "size": "1024x1024",
            "strength": 0.65,
        },
    )

    assert plan.effect is not None
    assert plan.effect.type == EffectType.LLM_CALL
    payload = dict(plan.effect.payload or {})
    output = dict(payload.get("output") or {})
    media = list(payload.get("media") or [])
    assert payload["prompt"] == "Make the jacket red."
    assert output["modality"] == "image"
    assert output["task"] == "image_edit"
    assert output["provider"] == "openai-compatible"
    assert output["model"] == "gpt-image-1"
    assert output["size"] == "1024x1024"
    assert output["strength"] == 0.65
    assert media[0]["$artifact"] == "source-img"
    assert media[0]["type"] == "image"
    assert media[0]["role"] == "source"
    assert media[1]["$artifact"] == "mask-img"
    assert media[1]["role"] == "mask"


def test_image_to_image_alias_compiles_to_image_edit_selector() -> None:
    plan = _plan_for_node(
        "image_to_image",
        {"prompt": "Add clouds.", "source_image": "img-1"},
    )

    assert plan.effect is not None
    payload = dict(plan.effect.payload or {})
    output = dict(payload.get("output") or {})
    assert output["task"] == "image_edit"
    assert payload["media"] == [{"type": "image", "role": "source", "$artifact": "img-1"}]


def test_generate_voice_node_compiles_to_llm_call_tts_selector() -> None:
    plan = _plan_for_node(
        "generate_voice",
        {
            "text": "hello",
            "voice": "clone:laurent",
            "profile": "piper/en_US",
            "tts_model": "en_US-lessac-medium.onnx",
            "format": "wav",
            "speed": 1.1,
        },
    )

    assert plan.effect is not None
    assert plan.effect.type == EffectType.LLM_CALL
    payload = dict(plan.effect.payload or {})
    output = dict(payload.get("output") or {})
    assert payload["prompt"] == "hello"
    assert "model" not in payload
    assert output["modality"] == "voice"
    assert output["task"] == "tts"
    assert output["voice"] == "clone:laurent"
    assert output["profile"] == "piper/en_US"
    assert output["model"] == "en_US-lessac-medium.onnx"
    assert output["format"] == "wav"
    assert output["speed"] == 1.1


def test_generate_music_node_compiles_to_llm_call_music_selector() -> None:
    plan = _plan_for_node(
        "generate_music",
        {
            "prompt": "Warm lo-fi piano with brushed drums.",
            "music_provider": "acemusic",
            "music_model": "ace-step",
            "music_backend": "acemusic",
            "format": "wav",
            "duration_s": 12,
            "seed": 7,
            "num_inference_steps": 20,
            "guidance_scale": 1.5,
            "instrumental": True,
            "lyrics": "no lyrics",
            "vocal_language": "en",
            "negative_prompt": "distortion",
            "sample_rate": 44100,
            "bpm": 92,
            "keyscale": "C minor",
            "timesignature": "4/4",
            "composition_plan": {"sections": ["intro", "loop"]},
            "positive_styles": ["lo-fi", "piano"],
            "negative_styles": ["metal"],
            "planning": False,
        },
    )

    assert plan.effect is not None
    assert plan.effect.type == EffectType.LLM_CALL
    payload = dict(plan.effect.payload or {})
    output = dict(payload.get("output") or {})
    assert payload["prompt"] == "Warm lo-fi piano with brushed drums."
    assert "provider" not in payload
    assert "model" not in payload
    assert output["modality"] == "music"
    assert output["task"] == "music_generation"
    assert output["format"] == "wav"
    assert output["provider"] == "acemusic"
    assert output["model"] == "ace-step"
    assert output["backend"] == "acemusic"
    assert output["duration_s"] == 12.0
    assert output["seed"] == 7
    assert output["num_inference_steps"] == 20
    assert output["guidance_scale"] == 1.5
    assert output["instrumental"] is True
    assert output["lyrics"] == "no lyrics"
    assert output["vocal_language"] == "en"
    assert output["negative_prompt"] == "distortion"
    assert output["sample_rate"] == 44100
    assert output["bpm"] == 92.0
    assert output["keyscale"] == "C minor"
    assert output["timesignature"] == "4/4"
    assert output["composition_plan"] == {"sections": ["intro", "loop"]}
    assert output["positive_styles"] == ["lo-fi", "piano"]
    assert output["negative_styles"] == ["metal"]
    assert output["planning"] is False


def test_generate_music_legacy_provider_model_are_not_media_fallbacks() -> None:
    plan = _plan_for_node(
        "generate_music",
        {
            "prompt": "a beat",
            "provider": "acemusic",
            "model": "ace-step",
        },
    )

    assert plan.effect is not None
    payload = dict(plan.effect.payload or {})
    output = dict(payload.get("output") or {})
    assert "provider" not in payload
    assert "model" not in payload
    assert "provider" not in output
    assert "model" not in output


def test_generate_music_does_not_treat_runtime_provider_input_as_music_provider() -> None:
    plan = _plan_for_node(
        "generate_music",
        {"prompt": "a beat", "music_model": "ace-step"},
        {"provider": "lmstudio", "model": "chat-model"},
    )

    assert plan.effect is not None
    payload = dict(plan.effect.payload or {})
    output = dict(payload.get("output") or {})
    assert "provider" not in output
    assert output["model"] == "ace-step"


def test_transcribe_audio_node_compiles_to_llm_call_media_selector() -> None:
    plan = _plan_for_node(
        "transcribe_audio",
        {
            "audio_artifact": {"$artifact": "audio-1", "content_type": "audio/wav"},
            "language": "en",
            "stt_model": "whisper-test",
            "format": "json",
            "temperature": 0,
        },
    )

    assert plan.effect is not None
    assert plan.effect.type == EffectType.LLM_CALL
    payload = dict(plan.effect.payload or {})
    output = dict(payload.get("output") or {})
    assert payload["media"] == [{"$artifact": "audio-1", "content_type": "audio/wav"}]
    assert "model" not in payload
    assert output["modality"] == "text"
    assert output["task"] == "transcription"
    assert output["language"] == "en"
    assert output["model"] == "whisper-test"
    assert output["format"] == "json"
    assert output["temperature"] == 0.0


def test_listen_voice_node_compiles_to_voice_wait_event() -> None:
    plan = _plan_for_node(
        "listen_voice",
        {
            "prompt": "Say your answer",
            "wait_key": "voice_answer",
            "language": "fr",
            "stt_provider": "openai",
            "stt_model": "whisper-test",
            "max_duration_s": 12,
        },
    )

    assert plan.effect is not None
    assert plan.effect.type == EffectType.WAIT_EVENT
    payload = dict(plan.effect.payload or {})
    details = dict(payload.get("details") or {})
    assert payload["wait_key"] == "voice_answer"
    assert payload["prompt"] == "Say your answer"
    assert payload["allow_free_text"] is True
    assert details["input_mode"] == "voice"
    assert details["language"] == "fr"
    assert details["provider"] == "openai"
    assert details["model"] == "whisper-test"
    assert details["max_duration_s"] == 12.0



from __future__ import annotations

import ast
import os
import subprocess
import sys
from pathlib import Path


ROOT = Path(__file__).resolve().parents[1]
SRC = ROOT / "src"
KERNEL_DIRS = (
    SRC / "abstractruntime" / "core",
    SRC / "abstractruntime" / "storage",
    SRC / "abstractruntime" / "scheduler",
    SRC / "abstractruntime" / "evidence",
    SRC / "abstractruntime" / "identity",
    SRC / "abstractruntime" / "rendering",
    SRC / "abstractruntime" / "workflow_bundle",
)
OPTIONAL_STACK_ROOTS = {"abstractcore", "abstractvision", "abstractvoice", "abstractmemory", "abstractmusic"}


def _import_root(module: str) -> str:
    return module.split(".", 1)[0]


def test_runtime_kernel_does_not_import_optional_capability_stacks() -> None:
    violations: list[str] = []
    for base in KERNEL_DIRS:
        for path in sorted(base.rglob("*.py")):
            tree = ast.parse(path.read_text(encoding="utf-8"), filename=str(path))
            for node in ast.walk(tree):
                if isinstance(node, ast.Import):
                    for alias in node.names:
                        if _import_root(alias.name) in OPTIONAL_STACK_ROOTS:
                            violations.append(f"{path.relative_to(ROOT)} imports {alias.name}")
                elif isinstance(node, ast.ImportFrom) and node.module:
                    if _import_root(node.module) in OPTIONAL_STACK_ROOTS:
                        violations.append(f"{path.relative_to(ROOT)} imports {node.module}")

    assert violations == []


def test_package_root_import_does_not_touch_optional_capability_stacks() -> None:
    script = f"""
import importlib.abc
import sys

sys.path.insert(0, {str(SRC)!r})

class OptionalStackBlocker(importlib.abc.MetaPathFinder):
    blocked = {sorted(OPTIONAL_STACK_ROOTS)!r}

    def find_spec(self, fullname, path=None, target=None):
        if fullname.split(".", 1)[0] in self.blocked:
            raise AssertionError(f"unexpected optional import: {{fullname}}")
        return None

sys.meta_path.insert(0, OptionalStackBlocker())
import abstractruntime
print(abstractruntime.Runtime.__name__)
"""
    env = dict(os.environ)
    env["PYTHONPATH"] = str(SRC)
    proc = subprocess.run(
        [sys.executable, "-c", script],
        cwd=str(ROOT),
        env=env,
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=False,
    )

    assert proc.returncode == 0, proc.stderr
    assert proc.stdout.strip() == "Runtime"



from __future__ import annotations

from pathlib import Path

import pytest

pytest.importorskip("abstractflow")
pytest.importorskip("abstractmemory")
pytest.importorskip("lancedb")

from abstractmemory import LanceDBTripleStore
from abstractruntime.core.runtime import Runtime
from abstractruntime.integrations.abstractmemory.effect_handlers import build_memory_kg_effect_handlers
from abstractruntime.storage.in_memory import InMemoryLedgerStore, InMemoryRunStore
from abstractruntime.visualflow_compiler.visual.executor import visual_to_flow
from abstractruntime.visualflow_compiler.visual.models import load_visualflow_json

from abstractflow import FlowRunner


class _DummyEmbedder:
    """Deterministic tiny embedder for tests (no external services)."""

    def embed_texts(self, texts):
        out = []
        for t in texts:
            t2 = str(t or "").lower()
            out.append([1.0 if "android" in t2 else 0.0, 0.1, 0.2])
        return out


def test_visual_memory_kg_query_propagates_active_memory_outputs(tmp_path) -> None:
    store_dir = Path(tmp_path) / "kg"
    store = LanceDBTripleStore(store_dir, embedder=_DummyEmbedder())

    run_store = InMemoryRunStore()
    runtime = Runtime(
        run_store=run_store,
        ledger_store=InMemoryLedgerStore(),
        effect_handlers=build_memory_kg_effect_handlers(store=store, run_store=run_store, now_iso=lambda: "2026-01-01T00:00:02Z"),
    )

    vf = load_visualflow_json(
        {
            "id": "test-kg-query-propagates-packing",
            "name": "test-kg-query-propagates-packing",
            "description": "",
            "interfaces": [],
            "nodes": [
                {
                    "id": "node-1",
                    "type": "on_flow_start",
                    "position": {"x": 0, "y": 0},
                    "data": {
                        "nodeType": "on_flow_start",
                        "label": "Start",
                        "inputs": [],
                        "outputs": [{"id": "exec-out", "label": "", "type": "execution"}],
                    },
                },
                {
                    "id": "node-2",
                    "type": "literal_array",
                    "position": {"x": 0, "y": 0},
                    "data": {
                        "nodeType": "literal_array",
                        "label": "Assertions",
                        "inputs": [],
                        "outputs": [{"id": "value", "label": "assertions", "type": "assertions"}],
                        "literalValue": [
                            {
                                "subject": "Data",
                                "predicate": "dcterms:description",
                                "object": "Android officer",
                                "observed_at": "2026-01-01T00:00:00Z",
                                "attributes": {"evidence_quote": "android"},
                                "provenance": {"span_id": "s1"},
                            }
                        ],
                    },
                },
                {
                    "id": "node-3",
                    "type": "memory_kg_assert",
                    "position": {"x": 0, "y": 0},
                    "data": {
                        "nodeType": "memory_kg_assert",
                        "label": "Assert",
                        "inputs": [
                            {"id": "exec-in", "label": "", "type": "execution"},
                            {"id": "assertions", "label": "assertions", "type": "assertions"},
                            {"id": "scope", "label": "scope", "type": "string"},
                            {"id": "owner_id", "label": "owner_id", "type": "string"},
                        ],
                        "outputs": [
                            {"id": "exec-out", "label": "", "type": "execution"},
                            {"id": "count", "label": "count", "type": "number"},
                            {"id": "ok", "label": "ok", "type": "boolean"},
                        ],
                        "pinDefaults": {"scope": "global", "owner_id": "global_memory"},
                    },
                },
                {
                    "id": "node-4",
                    "type": "memory_kg_query",
                    "position": {"x": 0, "y": 0},
                    "data": {
                        "nodeType": "memory_kg_query",
                        "label": "Query",
                        "inputs": [
                            {"id": "exec-in", "label": "", "type": "execution"},
                            {"id": "query_text", "label": "query_text", "type": "string"},
                            {"id": "scope", "label": "scope", "type": "string"},
                            {"id": "owner_id", "label": "owner_id", "type": "string"},
                            {"id": "min_score", "label": "min_score", "type": "number"},
                            {"id": "limit", "label": "limit", "type": "number"},
                            {"id": "max_input_tokens", "label": "max_input_tokens", "type": "number"},
                            {"id": "model", "label": "model", "type": "model"},
                        ],
                        "outputs": [
                            {"id": "exec-out", "label": "", "type": "execution"},
                            {"id": "active_memory_text", "label": "active_memory_text", "type": "string"},
                            {"id": "estimated_tokens", "label": "estimated_tokens", "type": "number"},
                            {"id": "count", "label": "count", "type": "number"},
                        ],
                        "pinDefaults": {
                            "query_text": "android",
                            "scope": "global",
                            "owner_id": "global_memory",
                            "min_score": 0.0,
                            "limit": 10,
                            "max_input_tokens": 80,
                            "model": "unit-test-model",
                        },
                    },
                },
                {
                    "id": "node-5",
                    "type": "on_flow_end",
                    "position": {"x": 0, "y": 0},
                    "data": {
                        "nodeType": "on_flow_end",
                        "label": "End",
                        "inputs": [
                            {"id": "exec-in", "label": "", "type": "execution"},
                            {"id": "active_memory_text", "label": "active_memory_text", "type": "string"},
                            {"id": "estimated_tokens", "label": "estimated_tokens", "type": "number"},
                            {"id": "count", "label": "count", "type": "number"},
                        ],
                        "outputs": [],
                    },
                },
            ],
            "edges": [
                {"id": "e1", "source": "node-1", "sourceHandle": "exec-out", "target": "node-3", "targetHandle": "exec-in", "animated": True},
                {"id": "e2", "source": "node-2", "sourceHandle": "value", "target": "node-3", "targetHandle": "assertions", "animated": False},
                {"id": "e3", "source": "node-3", "sourceHandle": "exec-out", "target": "node-4", "targetHandle": "exec-in", "animated": True},
                {"id": "e4", "source": "node-4", "sourceHandle": "exec-out", "target": "node-5", "targetHandle": "exec-in", "animated": True},
                {"id": "e5", "source": "node-4", "sourceHandle": "active_memory_text", "target": "node-5", "targetHandle": "active_memory_text", "animated": False},
                {"id": "e6", "source": "node-4", "sourceHandle": "estimated_tokens", "target": "node-5", "targetHandle": "estimated_tokens", "animated": False},
                {"id": "e7", "source": "node-4", "sourceHandle": "count", "target": "node-5", "targetHandle": "count", "animated": False},
            ],
            "entryNode": "node-1",
        }
    )

    flow = visual_to_flow(vf)
    runner = FlowRunner(flow, runtime=runtime)
    out = runner.run({})

    assert out.get("success") is True
    result = out.get("result")
    assert isinstance(result, dict)
    assert isinstance(result.get("active_memory_text"), str)
    assert result["active_memory_text"].startswith("## KG ACTIVE MEMORY")
    assert int(result.get("estimated_tokens") or 0) > 0
    assert int(result.get("count") or 0) == 1



from __future__ import annotations

from typing import Any, Dict, List

from abstractruntime import Effect, EffectType, RunState
from abstractruntime.integrations.abstractcore.effect_handlers import make_llm_call_handler


def _make_validation_error() -> Exception:
    from pydantic import BaseModel, ValidationError

    class _TmpModel(BaseModel):
        x: int

    try:
        _TmpModel.model_validate({})
    except ValidationError as e:
        return e
    raise AssertionError("Expected ValidationError")


class _FailStructuredThenReturn:
    def __init__(self, *, content: str) -> None:
        self.calls: List[Dict[str, Any]] = []
        self._content = content

    def generate(self, **kwargs):
        self.calls.append(dict(kwargs))
        params = kwargs.get("params") if isinstance(kwargs, dict) else None
        params = params if isinstance(params, dict) else {}
        if "response_model" in params:
            raise _make_validation_error()
        return {"content": self._content, "metadata": {}}


def test_llm_call_structured_output_fallback_disabled_still_fails() -> None:
    llm = _FailStructuredThenReturn(content='{"choice":"neutral"}')
    handler = make_llm_call_handler(llm=llm)

    run = RunState.new(workflow_id="wf", entry_node="n1", vars={})
    effect = Effect(
        type=EffectType.LLM_CALL,
        payload={
            "prompt": "hi",
            "response_schema": {"type": "object", "properties": {"choice": {"type": "string"}}},
        },
    )

    outcome = handler(run, effect, None)
    assert outcome.status == "failed"
    assert len(llm.calls) == 1


def test_llm_call_structured_output_fallback_retries_and_parses() -> None:
    llm = _FailStructuredThenReturn(content='{"choice":"neutral"}')
    handler = make_llm_call_handler(llm=llm)

    run = RunState.new(workflow_id="wf", entry_node="n1", vars={})
    effect = Effect(
        type=EffectType.LLM_CALL,
        payload={
            "prompt": "hi",
            "response_schema": {"type": "object", "properties": {"choice": {"type": "string"}}},
            "structured_output_fallback": True,
        },
    )

    outcome = handler(run, effect, None)
    assert outcome.status == "completed"
    assert len(llm.calls) == 2

    assert isinstance(outcome.result, dict)
    assert outcome.result.get("data") == {"choice": "neutral"}
    meta = outcome.result.get("metadata")
    assert isinstance(meta, dict)
    assert isinstance(meta.get("_structured_output_fallback"), dict)
    assert meta["_structured_output_fallback"].get("used") is True



from __future__ import annotations

from abstractruntime import Runtime
from abstractruntime.core.models import RunStatus
from abstractruntime.storage.in_memory import InMemoryLedgerStore, InMemoryRunStore
from abstractruntime.visualflow_compiler import compile_visualflow


def test_visualflow_get_context_and_builder_nodes() -> None:
    flow = {
        "id": "context-and-builders",
        "name": "context-and-builders",
        "entryNode": "node-1",
        "nodes": [
            {
                "id": "node-1",
                "type": "on_flow_start",
                "data": {"inputs": [], "outputs": [{"id": "exec-out", "label": "", "type": "execution"}]},
            },
	            {
	                "id": "node-2",
	                "type": "make_context",
	                "data": {
	                    "inputs": [
	                        {"id": "task", "label": "task", "type": "string"},
	                        {"id": "messages", "label": "messages", "type": "array"},
	                        {"id": "context_extra", "label": "context_extra", "type": "object"},
	                    ],
	                    "outputs": [{"id": "context", "label": "context", "type": "object"}],
	                    "pinDefaults": {
	                        "task": "hello",
	                        "messages": [{"role": "user", "content": "hi"}],
	                        "context_extra": {"foo": 1},
	                    },
	                },
	            },
            {
                "id": "node-3",
                "type": "make_meta",
                "data": {
                    "inputs": [
                        {"id": "schema", "label": "schema", "type": "string"},
                        {"id": "version", "label": "version", "type": "number"},
                        {"id": "provider", "label": "provider", "type": "provider"},
                        {"id": "model", "label": "model", "type": "model"},
                        {"id": "usage", "label": "usage", "type": "object"},
                        {"id": "trace_id", "label": "trace_id", "type": "string"},
                        {"id": "warnings", "label": "warnings", "type": "array"},
                        {"id": "debug", "label": "debug", "type": "object"},
                        {"id": "extra", "label": "extra", "type": "object"},
                    ],
                    "outputs": [{"id": "meta", "label": "meta", "type": "object"}],
                    "pinDefaults": {
                        "provider": "lmstudio",
                        "model": "qwen/qwen3-next-80b",
                        "usage": {"input_tokens": 1, "output_tokens": 2},
                        "trace_id": "trace-123",
                        "warnings": ["w1"],
                        "debug": {"a": 1},
                        "extra": {"schema": "should_be_overridden"},
                    },
                },
            },
	            {
	                "id": "node-4",
	                "type": "make_scratchpad",
	                "data": {
	                    "inputs": [
	                        {"id": "sub_run_id", "label": "sub_run_id", "type": "string"},
	                        {"id": "workflow_id", "label": "workflow_id", "type": "string"},
	                        {"id": "task", "label": "task", "type": "string"},
	                        {"id": "messages", "label": "messages", "type": "array"},
	                        {"id": "context_extra", "label": "context_extra", "type": "object"},
	                        {"id": "node_traces", "label": "node_traces", "type": "object"},
	                        {"id": "steps", "label": "steps", "type": "array"},
	                        {"id": "tool_calls", "label": "tool_calls", "type": "array"},
	                        {"id": "tool_results", "label": "tool_results", "type": "array"},
	                    ],
	                    "outputs": [{"id": "scratchpad", "label": "scratchpad", "type": "object"}],
	                    "pinDefaults": {
	                        "sub_run_id": "sub-1",
	                        "workflow_id": "wf-1",
	                        "task": "scratch-task",
	                        "messages": [{"role": "user", "content": "hi"}],
	                        "context_extra": {"foo": "bar"},
	                        "node_traces": {"node-9": {"node_id": "node-9", "steps": []}},
	                        "steps": [{"type": "llm_call"}],
	                        "tool_calls": [],
	                        "tool_results": [],
	                    },
	                },
	            },
	            {
	                "id": "node-6",
	                "type": "get_context",
                "data": {
                    "inputs": [],
                    "outputs": [
                        {"id": "context", "label": "context", "type": "object"},
                        {"id": "task", "label": "task", "type": "string"},
                        {"id": "messages", "label": "messages", "type": "array"},
                    ],
                },
            },
            {
                "id": "node-7",
                "type": "on_flow_end",
	                "data": {
	                    "inputs": [
	                        {"id": "exec-in", "label": "", "type": "execution"},
	                        {"id": "built_context", "label": "built_context", "type": "object"},
	                        {"id": "built_meta", "label": "built_meta", "type": "object"},
	                        {"id": "built_scratchpad", "label": "built_scratchpad", "type": "object"},
	                        {"id": "got_context", "label": "got_context", "type": "object"},
	                        {"id": "got_task", "label": "got_task", "type": "string"},
	                        {"id": "got_messages", "label": "got_messages", "type": "array"},
	                    ],
                    "outputs": [],
                },
            },
        ],
	        "edges": [
	            {"id": "e-exec", "source": "node-1", "target": "node-7", "sourceHandle": "exec-out", "targetHandle": "exec-in"},
	            {"id": "e-bc", "source": "node-2", "target": "node-7", "sourceHandle": "context", "targetHandle": "built_context"},
	            {"id": "e-bm", "source": "node-3", "target": "node-7", "sourceHandle": "meta", "targetHandle": "built_meta"},
	            {"id": "e-bs", "source": "node-4", "target": "node-7", "sourceHandle": "scratchpad", "targetHandle": "built_scratchpad"},
	            {"id": "e-gc", "source": "node-6", "target": "node-7", "sourceHandle": "context", "targetHandle": "got_context"},
	            {"id": "e-gt", "source": "node-6", "target": "node-7", "sourceHandle": "task", "targetHandle": "got_task"},
	            {"id": "e-gm", "source": "node-6", "target": "node-7", "sourceHandle": "messages", "targetHandle": "got_messages"},
	        ],
    }

    spec = compile_visualflow(flow)
    runtime = Runtime(run_store=InMemoryRunStore(), ledger_store=InMemoryLedgerStore())

    run_vars = {"context": {"task": "from_run", "messages": [{"role": "user", "content": "hello"}], "other": True}}

    run_id = runtime.start(workflow=spec, vars=run_vars)
    state = runtime.tick(workflow=spec, run_id=run_id)

    assert state.status == RunStatus.COMPLETED
    assert isinstance(state.output, dict)
    assert state.output.get("success") is True
    assert "result" not in state.output

    result = state.output

    # Get Context
    assert result["got_context"]["task"] == "from_run"
    assert result["got_task"] == "from_run"
    assert result["got_messages"] == [{"role": "user", "content": "hello"}]

    # Make Context
    assert result["built_context"]["task"] == "hello"
    assert result["built_context"]["messages"] == [{"role": "user", "content": "hi"}]
    assert result["built_context"]["foo"] == 1

    # Make Meta (extra.schema overridden by default schema)
    assert result["built_meta"]["schema"] == "abstractcode.agent.v1.meta"
    assert result["built_meta"]["version"] == 1
    assert result["built_meta"]["provider"] == "lmstudio"
    assert result["built_meta"]["model"] == "qwen/qwen3-next-80b"
    assert result["built_meta"]["usage"] == {"input_tokens": 1, "output_tokens": 2}
    assert result["built_meta"]["trace"]["trace_id"] == "trace-123"
    assert result["built_meta"]["warnings"] == ["w1"]
    assert result["built_meta"]["debug"] == {"a": 1}

    # Make Scratchpad
    assert result["built_scratchpad"]["sub_run_id"] == "sub-1"
    assert result["built_scratchpad"]["workflow_id"] == "wf-1"
    assert result["built_scratchpad"]["task"] == "scratch-task"
    assert result["built_scratchpad"]["messages"] == [{"role": "user", "content": "hi"}]
    assert result["built_scratchpad"]["context_extra"]["foo"] == "bar"
    assert result["built_scratchpad"]["node_traces"]["node-9"]["node_id"] == "node-9"
    assert result["built_scratchpad"]["steps"][0]["type"] == "llm_call"



"""Regression: START_SUBWORKFLOW supports async+wait mode.

Goal:
- Allow hosts to start a child run without blocking the parent tick (async=True),
  but still keep the parent in a durable waiting state (wait=True) until the host
  resumes it with the child's final output.
"""

from __future__ import annotations

from abstractruntime import (
    Effect,
    EffectType,
    InMemoryLedgerStore,
    InMemoryRunStore,
    RunState,
    RunStatus,
    Runtime,
    StepPlan,
    WaitReason,
    WorkflowRegistry,
    WorkflowSpec,
)


def test_start_subworkflow_async_wait_puts_parent_in_subworkflow_wait() -> None:
    def child_node(run: RunState, ctx) -> StepPlan:
        return StepPlan(node_id="child", complete_output={"answer": "ok"})

    child = WorkflowSpec(workflow_id="child_wf", entry_node="child", nodes={"child": child_node})

    def parent_node(run: RunState, ctx) -> StepPlan:
        return StepPlan(
            node_id="parent",
            effect=Effect(
                type=EffectType.START_SUBWORKFLOW,
                payload={"workflow_id": "child_wf", "async": True, "wait": True},
                result_key="sub_result",
            ),
            next_node="after",
        )

    def after_node(run: RunState, ctx) -> StepPlan:
        return StepPlan(node_id="after", complete_output={"sub": run.vars.get("sub_result")})

    parent = WorkflowSpec(workflow_id="parent_wf", entry_node="parent", nodes={"parent": parent_node, "after": after_node})

    reg = WorkflowRegistry()
    reg.register(child)
    reg.register(parent)

    rt = Runtime(run_store=InMemoryRunStore(), ledger_store=InMemoryLedgerStore(), workflow_registry=reg)
    run_id = rt.start(workflow=parent, vars={})

    st = rt.tick(workflow=parent, run_id=run_id, max_steps=1)
    assert st.status == RunStatus.WAITING
    assert st.waiting is not None
    assert st.waiting.reason == WaitReason.SUBWORKFLOW
    assert isinstance(st.waiting.wait_key, str) and st.waiting.wait_key.startswith("subworkflow:")





from __future__ import annotations

import json
import sqlite3
import threading
import multiprocessing
from pathlib import Path

from abstractruntime.core.models import StepRecord, StepStatus
from abstractruntime.storage.sqlite import SqliteDatabase, SqliteLedgerStore


def _mp_sqlite_ledger_store_append_worker(
    *,
    db_path: str,
    run_id: str,
    worker_id: int,
    start_evt: multiprocessing.synchronize.Event,
    errors_q: multiprocessing.queues.Queue[str],
    count: int,
) -> None:
    try:
        if not start_evt.wait(timeout=5.0):
            raise RuntimeError("start timeout")
        db = SqliteDatabase(Path(db_path))
        store = SqliteLedgerStore(db)
        for j in range(int(count)):
            store.append(StepRecord(run_id=run_id, step_id=f"{worker_id}:{j}", node_id="n", status=StepStatus.COMPLETED))
    except BaseException as e:
        try:
            errors_q.put_nowait(repr(e))
        except Exception:
            pass


def test_sqlite_ledger_store_append_list_count_and_restart(tmp_path: Path) -> None:
    db_path = tmp_path / "gateway.sqlite3"
    db = SqliteDatabase(db_path)
    store = SqliteLedgerStore(db)

    run_id = "run_1"
    store.append(StepRecord(run_id=run_id, step_id="s1", node_id="n1", status=StepStatus.STARTED))
    store.append(StepRecord(run_id=run_id, step_id="s2", node_id="n2", status=StepStatus.COMPLETED))

    assert store.count(run_id) == 2
    items = store.list(run_id)
    assert [x.get("step_id") for x in items] == ["s1", "s2"]

    page1, cur1 = store.list_after(run_id=run_id, after=0, limit=1)
    assert [x.get("step_id") for x in page1] == ["s1"]
    assert cur1 == 1

    page2, cur2 = store.list_after(run_id=run_id, after=cur1, limit=10)
    assert [x.get("step_id") for x in page2] == ["s2"]
    assert cur2 == 2

    # Restart: re-open and verify count/list still work.
    db2 = SqliteDatabase(db_path)
    store2 = SqliteLedgerStore(db2)
    assert store2.count(run_id) == 2
    assert [x.get("step_id") for x in store2.list(run_id)] == ["s1", "s2"]


def test_sqlite_ledger_store_count_many_and_metrics_many(tmp_path: Path) -> None:
    db_path = tmp_path / "gateway.sqlite3"
    db = SqliteDatabase(db_path)
    store = SqliteLedgerStore(db)

    run_id = "run_1"
    # Started record shouldn't count toward metrics_many (completed-only).
    store.append(
        StepRecord(
            run_id=run_id,
            step_id="s0",
            node_id="n0",
            status=StepStatus.STARTED,
            effect={"type": "llm_call", "payload": {}, "result_key": "_tmp"},
        )
    )
    # Completed LLM call with usage.
    store.append(
        StepRecord(
            run_id=run_id,
            step_id="s1",
            node_id="n1",
            status=StepStatus.COMPLETED,
            effect={"type": "llm_call", "payload": {}, "result_key": "_tmp"},
            result={"usage": {"input_tokens": 10, "output_tokens": 2, "total_tokens": 12}},
        )
    )
    # Completed tool_calls with 2 tool calls.
    store.append(
        StepRecord(
            run_id=run_id,
            step_id="s2",
            node_id="n2",
            status=StepStatus.COMPLETED,
            effect={"type": "tool_calls", "payload": {"tool_calls": [{"name": "a"}, {"name": "b"}]}, "result_key": "_tmp"},
            result={},
        )
    )

    assert store.count_many([run_id, "missing"]) == {run_id: 3}

    metrics = store.metrics_many([run_id, "missing"])
    assert metrics.get(run_id) == {"steps": 2, "llm_calls": 1, "tool_calls": 2, "tokens_total": 12}


def test_sqlite_ledger_store_append_is_concurrency_safe(tmp_path: Path) -> None:
    db_path = tmp_path / "gateway.sqlite3"
    db = SqliteDatabase(db_path)
    store = SqliteLedgerStore(db)

    run_id = "run_1"

    errors: list[BaseException] = []
    barrier = threading.Barrier(8)

    def _worker(worker_id: int) -> None:
        try:
            barrier.wait(timeout=5.0)
            for j in range(50):
                store.append(StepRecord(run_id=run_id, step_id=f"{worker_id}:{j}", node_id="n", status=StepStatus.COMPLETED))
        except BaseException as e:
            errors.append(e)

    threads = [threading.Thread(target=_worker, args=(i,), daemon=True) for i in range(8)]
    for t in threads:
        t.start()
    for t in threads:
        t.join(timeout=10.0)

    assert not errors
    assert store.count(run_id) == 400


def test_sqlite_ledger_store_append_is_multiprocess_concurrency_safe(tmp_path: Path) -> None:
    db_path = tmp_path / "gateway.sqlite3"
    run_id = "run_1"

    # Use a multiprocess test to approximate real gateway deployments where
    # API + runner may append concurrently from different processes.
    ctx = multiprocessing.get_context("spawn")
    start_evt = ctx.Event()
    errors_q: multiprocessing.queues.Queue[str] = ctx.Queue()
    procs = [
        ctx.Process(
            target=_mp_sqlite_ledger_store_append_worker,
            kwargs={
                "db_path": str(db_path),
                "run_id": run_id,
                "worker_id": i,
                "start_evt": start_evt,
                "errors_q": errors_q,
                "count": 40,
            },
            daemon=True,
        )
        for i in range(4)
    ]
    for p in procs:
        p.start()
    start_evt.set()
    for p in procs:
        p.join(timeout=20.0)

    errors: list[str] = []
    try:
        while True:
            errors.append(errors_q.get_nowait())
    except Exception:
        pass

    for p in procs:
        if p.is_alive():
            p.terminate()
        assert p.exitcode == 0

    assert not errors
    db2 = SqliteDatabase(db_path)
    store2 = SqliteLedgerStore(db2)
    assert store2.count(run_id) == 160


def test_sqlite_ledger_store_backfills_heads_from_existing_ledger(tmp_path: Path) -> None:
    db_path = tmp_path / "gateway.sqlite3"

    # Simulate an older DB: ledger exists and contains records, but ledger_heads does not.
    conn = sqlite3.connect(str(db_path))
    try:
        conn.execute(
            """
            CREATE TABLE ledger (
              run_id TEXT NOT NULL,
              seq INTEGER NOT NULL,
              record_json TEXT NOT NULL,
              PRIMARY KEY (run_id, seq)
            );
            """
        )
        conn.execute("INSERT INTO ledger (run_id, seq, record_json) VALUES (?, ?, ?);", ("run_1", 1, json.dumps({"step_id": "s1"})))
        conn.execute("INSERT INTO ledger (run_id, seq, record_json) VALUES (?, ?, ?);", ("run_1", 2, json.dumps({"step_id": "s2"})))
        conn.commit()
    finally:
        conn.close()

    # New store must continue from seq=2 (so next append is seq=3).
    db = SqliteDatabase(db_path)
    store = SqliteLedgerStore(db)
    store.append(StepRecord(run_id="run_1", step_id="s3", node_id="n3", status=StepStatus.COMPLETED))

    assert store.count("run_1") == 3

    row = db.connection().execute("SELECT last_seq FROM ledger_heads WHERE run_id = ?;", ("run_1",)).fetchone()
    assert row is not None
    assert int(row["last_seq"] or 0) == 3


def test_sqlite_ledger_store_append_duplicate_seq_error_includes_context(tmp_path: Path) -> None:
    db_path = tmp_path / "gateway.sqlite3"
    db = SqliteDatabase(db_path)
    store = SqliteLedgerStore(db)

    run_id = "run_1"
    store.append(StepRecord(run_id=run_id, step_id="s1", node_id="n1", status=StepStatus.STARTED))

    # Corrupt the head to simulate a broken/partial migration or a writer bug.
    conn = db.connection()
    with conn:
        conn.execute("UPDATE ledger_heads SET last_seq = 0 WHERE run_id = ?;", (run_id,))

    try:
        store.append(StepRecord(run_id=run_id, step_id="s2", node_id="n2", status=StepStatus.COMPLETED))
        raise AssertionError("expected append to fail due to duplicate seq allocation")
    except sqlite3.IntegrityError as e:
        msg = str(e)
        assert "UNIQUE constraint failed: ledger.run_id, ledger.seq" in msg
        assert "run_id='run_1'" in msg
        assert "seq=1" in msg



from abstractruntime import Effect, EffectType, RunState
from abstractruntime.integrations.abstractcore.effect_handlers import make_llm_call_handler


class _StubLLM:
    def generate(self, **kwargs):
        return {"content": "ok", "metadata": {}}


def test_llm_call_handler_attaches_runtime_observability_payload() -> None:
    run = RunState.new(workflow_id="wf", entry_node="n1", vars={})
    effect = Effect(
        type=EffectType.LLM_CALL,
        payload={"prompt": "hello", "params": {"temperature": 0.2}},
    )

    handler = make_llm_call_handler(llm=_StubLLM())
    outcome = handler(run, effect, None)

    assert outcome.status == "completed"
    assert isinstance(outcome.result, dict)
    meta = outcome.result.get("metadata")
    assert isinstance(meta, dict)
    obs = meta.get("_runtime_observability")
    assert isinstance(obs, dict)
    captured = obs.get("llm_generate_kwargs")
    assert isinstance(captured, dict)
    assert captured["prompt"] == "hello"
    assert isinstance(captured.get("params"), dict)
    assert "trace_metadata" in captured["params"]




from __future__ import annotations

from typing import Any, Dict, List, Optional

from abstractruntime import InMemoryLedgerStore, InMemoryRunStore, RunState, RunStatus, Runtime, WorkflowSpec
from abstractruntime.core.models import Effect, EffectType, StepPlan
from abstractruntime.integrations.abstractcore.effect_handlers import make_llm_call_handler
from abstractruntime.integrations.abstractcore.llm_client import HttpResponse, RemoteAbstractCoreLLMClient, _normalize_local_response


class _StubLLM:
    def __init__(self) -> None:
        self.calls: List[Dict[str, Any]] = []

    def generate(
        self,
        *,
        prompt: str,
        messages: Optional[List[Dict[str, str]]] = None,
        system_prompt: Optional[str] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        media: Optional[List[Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        self.calls.append(
            {
                "prompt": prompt,
                "messages": messages,
                "system_prompt": system_prompt,
                "tools": tools,
                "media": media,
                "params": dict(params or {}),
            }
        )
        return {"content": "ok", "trace_id": "tr_test", "metadata": {"trace_id": "tr_test"}}


def test_llm_call_injects_trace_metadata() -> None:
    stub = _StubLLM()
    runtime = Runtime(
        run_store=InMemoryRunStore(),
        ledger_store=InMemoryLedgerStore(),
        effect_handlers={EffectType.LLM_CALL: make_llm_call_handler(llm=stub)},
    )

    def ask(run: RunState, ctx) -> StepPlan:
        return StepPlan(
            node_id="ask",
            effect=Effect(
                type=EffectType.LLM_CALL,
                payload={"prompt": "hello", "params": {"max_tokens": 5}},
                result_key="llm",
            ),
            next_node="done",
        )

    def done(run: RunState, ctx) -> StepPlan:
        return StepPlan(node_id="done", complete_output={"llm": run.vars.get("llm")})

    workflow = WorkflowSpec(workflow_id="trace_inject", entry_node="ask", nodes={"ask": ask, "done": done})

    run_id = runtime.start(workflow=workflow, actor_id="ar_test", session_id="sess_test")
    state = runtime.tick(workflow=workflow, run_id=run_id)

    assert state.status == RunStatus.COMPLETED
    assert stub.calls, "Expected stub LLM to be called"
    trace = (stub.calls[-1]["params"].get("trace_metadata") or {})
    assert trace.get("run_id") == run_id
    assert trace.get("actor_id") == "ar_test"
    assert trace.get("session_id") == "sess_test"


def test_normalize_local_response_extracts_trace_id() -> None:
    from abstractcore.core.types import GenerateResponse

    resp = GenerateResponse(content="hi", metadata={"trace_id": "tr_123"})
    out = _normalize_local_response(resp)
    assert out.get("trace_id") == "tr_123"
    assert isinstance(out.get("metadata"), dict)
    assert out["metadata"].get("trace_id") == "tr_123"


def test_normalize_local_response_extracts_reasoning_content() -> None:
    from abstractcore.core.types import GenerateResponse

    resp = GenerateResponse(
        content="\n\n",
        raw_response={
            "choices": [
                {
                    "message": {
                        "role": "assistant",
                        "content": "\n\n",
                        "reasoning_content": "I will call tools now.",
                        "tool_calls": [{"type": "function", "id": "1", "function": {"name": "list_files", "arguments": "{}"}}],
                    }
                }
            ]
        },
    )
    out = _normalize_local_response(resp)
    assert out.get("reasoning") == "I will call tools now."


def test_remote_client_sends_trace_headers_and_reads_trace_id() -> None:
    captured: Dict[str, Any] = {}

    class _Sender:
        def post(self, url: str, *, headers: Dict[str, str], json: Dict[str, Any], timeout: float) -> HttpResponse:
            captured["url"] = url
            captured["headers"] = dict(headers)
            captured["json"] = dict(json)
            return HttpResponse(
                body={
                    "model": "m",
                    "choices": [{"message": {"content": "ok"}, "finish_reason": "stop"}],
                },
                headers={"X-AbstractCore-Trace-Id": "tr_remote"},
            )

    client = RemoteAbstractCoreLLMClient(
        server_base_url="http://example.test",
        model="m",
        headers={"User-Agent": "test"},
        request_sender=_Sender(),
    )

    out = client.generate(prompt="hi", params={"trace_metadata": {"run_id": "r1", "actor_id": "a1"}})
    assert out.get("trace_id") == "tr_remote"
    assert captured["headers"].get("X-AbstractCore-Trace-Metadata")
    assert captured["headers"].get("X-AbstractCore-Run-Id") == "r1"
    assert captured["headers"].get("X-AbstractCore-Actor-Id") == "a1"



from __future__ import annotations

from abstractruntime.integrations.abstractcore.tool_executor import MappingToolExecutor


def _echo_file_path(*, file_path: str) -> str:
    return file_path


def _echo_search(*, pattern: str, path: str = ".") -> str:
    return f"{pattern}@{path}"


def test_mapping_tool_executor_normalizes_keys_by_separators_and_case() -> None:
    exe = MappingToolExecutor.from_tools([_echo_file_path])
    result = exe.execute(
        tool_calls=[
            {
                "name": "_echo_file_path",
                "arguments": {"filePath": "demo.txt"},
                "call_id": "1",
            }
        ]
    )
    assert result["results"][0]["success"] is True
    assert result["results"][0]["output"] == "demo.txt"


def test_mapping_tool_executor_applies_small_synonym_table() -> None:
    exe = MappingToolExecutor.from_tools([_echo_search])
    result = exe.execute(
        tool_calls=[
            {
                "name": "_echo_search",
                "arguments": {"query": "pygame", "directory": "/tmp"},
                "call_id": "1",
            }
        ]
    )
    assert result["results"][0]["success"] is True
    assert result["results"][0]["output"] == "pygame@/tmp"


def test_mapping_tool_executor_parses_json_string_arguments() -> None:
    exe = MappingToolExecutor.from_tools([_echo_file_path])
    result = exe.execute(
        tool_calls=[
            {
                "name": "_echo_file_path",
                "arguments": "{\"file-path\": \"demo.txt\"}",
                "call_id": "1",
            }
        ]
    )
    assert result["results"][0]["success"] is True
    assert result["results"][0]["output"] == "demo.txt"





from __future__ import annotations

from pathlib import Path

import pytest


pytestmark = pytest.mark.integration


def test_default_tool_specs_include_skim_files_and_executor_runs(tmp_path: Path) -> None:
    from abstractruntime.integrations.abstractcore.default_tools import get_default_tools, list_default_tool_specs
    from abstractruntime.integrations.abstractcore.tool_executor import MappingToolExecutor

    specs = list_default_tool_specs()
    names = {s.get("name") for s in specs if isinstance(s, dict)}
    assert "skim_files" in names

    skim = next((s for s in specs if isinstance(s, dict) and s.get("name") == "skim_files"), None)
    assert isinstance(skim, dict)
    params = skim.get("parameters") or {}
    assert isinstance(params, dict)
    paths_schema = params.get("paths") or {}
    assert isinstance(paths_schema, dict)
    assert paths_schema.get("type") == "array"
    assert "max_output_lines" not in params
    assert "max_chars_per_excerpt" not in params

    target = tmp_path / "demo.txt"
    target.write_text("one.\n\ntwo.\n\nthree.\n", encoding="utf-8")

    executor = MappingToolExecutor.from_tools(get_default_tools())
    result = executor.execute(
        tool_calls=[
            {
                "call_id": "1",
                "name": "skim_files",
                "arguments": {"paths": [str(target)], "target_percent": 8.0},
            }
        ]
    )

    assert isinstance(result, dict)
    results = result.get("results")
    assert isinstance(results, list) and results
    first = results[0]
    assert isinstance(first, dict)
    assert first.get("success") is True
    out = first.get("output")
    assert isinstance(out, str)
    assert "File:" in out
    assert "1:" in out



"""WorkflowBundle reader (directory or .flow zip).

This module focuses on *reading* bundles. Writing/packing bundles is expected to
be performed by authoring tooling (e.g., AbstractFlow).
"""

from __future__ import annotations

import json
import zipfile
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional

from .models import WorkflowBundleError, WorkflowBundleManifest, workflow_bundle_manifest_from_dict


@dataclass(frozen=True)
class WorkflowBundle:
    """An opened WorkflowBundle source."""

    source: Path
    manifest: WorkflowBundleManifest

    def _is_zip(self) -> bool:
        return self.source.is_file()

    def read_bytes(self, relpath: str) -> bytes:
        p = str(relpath or "").strip()
        if not p:
            raise WorkflowBundleError("read_bytes requires a non-empty relpath")

        if self._is_zip():
            with zipfile.ZipFile(self.source, "r") as zf:
                try:
                    return zf.read(p)
                except KeyError as e:
                    raise FileNotFoundError(f"Bundle file not found: {p}") from e
        else:
            abs_p = (self.source / p).resolve()
            # Ensure relpath cannot escape the bundle directory.
            try:
                abs_p.relative_to(self.source.resolve())
            except Exception as e:
                raise WorkflowBundleError(f"Unsafe relpath outside bundle dir: {p}") from e
            if not abs_p.exists():
                raise FileNotFoundError(f"Bundle file not found: {p}")
            return abs_p.read_bytes()

    def read_text(self, relpath: str, *, encoding: str = "utf-8") -> str:
        return self.read_bytes(relpath).decode(encoding)

    def read_json(self, relpath: str) -> Any:
        return json.loads(self.read_text(relpath))


def _read_manifest_from_dir(dir_path: Path) -> WorkflowBundleManifest:
    p = (dir_path / "manifest.json").resolve()
    if not p.exists():
        raise FileNotFoundError(f"manifest.json not found in bundle dir: {dir_path}")
    raw = json.loads(p.read_text(encoding="utf-8"))
    return workflow_bundle_manifest_from_dict(raw)


def _read_manifest_from_zip(zip_path: Path) -> WorkflowBundleManifest:
    with zipfile.ZipFile(zip_path, "r") as zf:
        try:
            raw_bytes = zf.read("manifest.json")
        except KeyError as e:
            raise FileNotFoundError(f"manifest.json not found in bundle: {zip_path}") from e
    raw = json.loads(raw_bytes.decode("utf-8"))
    return workflow_bundle_manifest_from_dict(raw)


def open_workflow_bundle(source: str | Path) -> WorkflowBundle:
    """Open a WorkflowBundle from a directory or a `.flow` zip file."""
    p = Path(source).expanduser().resolve()
    if not p.exists():
        raise FileNotFoundError(f"Bundle source not found: {p}")
    if p.is_dir():
        manifest = _read_manifest_from_dir(p)
        return WorkflowBundle(source=p, manifest=manifest)
    # File: treat as zip bundle.
    manifest = _read_manifest_from_zip(p)
    return WorkflowBundle(source=p, manifest=manifest)





from __future__ import annotations

import json
from typing import Any, Dict

from abstractruntime import Effect, EffectType, Runtime, StepPlan, WorkflowSpec
from abstractruntime.core.models import RunStatus
from abstractruntime.storage.in_memory import InMemoryLedgerStore, InMemoryRunStore


def test_vars_query_effect_reads_scratchpad_paths_and_keys() -> None:
    run_store = InMemoryRunStore()
    ledger_store = InMemoryLedgerStore()
    runtime = Runtime(run_store=run_store, ledger_store=ledger_store)

    vars: Dict[str, Any] = {
        "context": {"task": "t", "messages": []},
        "scratchpad": {"foo": {"bar": [1, 2, 3]}},
        "_runtime": {},
        "_temp": {},
        "_limits": {},
    }

    def inspect_value(run, ctx) -> StepPlan:
        return StepPlan(
            node_id="inspect_value",
            effect=Effect(
                type=EffectType.VARS_QUERY,
                payload={"path": "scratchpad.foo.bar[1]"},
                result_key="_temp.value",
            ),
            next_node="inspect_keys",
        )

    def inspect_keys(run, ctx) -> StepPlan:
        return StepPlan(
            node_id="inspect_keys",
            effect=Effect(
                type=EffectType.VARS_QUERY,
                payload={"path": "scratchpad.foo", "keys_only": True},
                result_key="_temp.keys",
            ),
            next_node="inspect_missing",
        )

    def inspect_missing(run, ctx) -> StepPlan:
        return StepPlan(
            node_id="inspect_missing",
            effect=Effect(
                type=EffectType.VARS_QUERY,
                payload={"path": "scratchpad.nope"},
                result_key="_temp.missing",
            ),
            next_node="done",
        )

    def done(run, ctx) -> StepPlan:
        temp = run.vars.get("_temp") if isinstance(run.vars.get("_temp"), dict) else {}
        return StepPlan(
            node_id="done",
            complete_output={
                "value": temp.get("value"),
                "keys": temp.get("keys"),
                "missing": temp.get("missing"),
            },
        )

    wf = WorkflowSpec(
        workflow_id="wf_vars_query",
        entry_node="inspect_value",
        nodes={
            "inspect_value": inspect_value,
            "inspect_keys": inspect_keys,
            "inspect_missing": inspect_missing,
            "done": done,
        },
    )

    run_id = runtime.start(workflow=wf, vars=vars)
    state = runtime.tick(workflow=wf, run_id=run_id)

    assert state.status == RunStatus.COMPLETED
    out = state.output or {}

    value = out.get("value")
    assert isinstance(value, dict)
    v_results = value.get("results")
    assert isinstance(v_results, list) and v_results
    v_text = str(v_results[0].get("output") or "")
    v_payload = json.loads(v_text)
    assert v_payload["path"] == "scratchpad.foo.bar[1]"
    assert v_payload["value"] == 2

    keys = out.get("keys")
    assert isinstance(keys, dict)
    k_results = keys.get("results")
    assert isinstance(k_results, list) and k_results
    k_text = str(k_results[0].get("output") or "")
    k_payload = json.loads(k_text)
    assert k_payload["path"] == "scratchpad.foo"
    assert "bar" in (k_payload.get("keys") or [])

    missing = out.get("missing")
    assert isinstance(missing, dict)
    m_results = missing.get("results")
    assert isinstance(m_results, list) and m_results
    assert m_results[0].get("success") is False
    assert "Missing key" in str(m_results[0].get("error") or "")




from __future__ import annotations

from abstractruntime.visualflow_compiler.visual.builtins import get_builtin_handler


def test_string_contains_builtin() -> None:
    contains = get_builtin_handler("contains")
    assert contains is not None

    assert contains({"text": "hello world", "pattern": "world"}) is True
    assert contains({"text": "hello world", "pattern": "WORLD"}) is False

    # Safer than Python's default behavior where "" is contained in every string.
    assert contains({"text": "hello", "pattern": ""}) is False
    assert contains({"text": "hello", "pattern": None}) is False
    assert contains({"text": None, "pattern": "x"}) is False


def test_string_replace_builtin_modes() -> None:
    replace = get_builtin_handler("replace")
    assert replace is not None

    assert replace({"text": "a-b-c", "pattern": "-", "replacement": "_", "mode": "all"}) == "a_b_c"
    assert replace({"text": "a-b-c", "pattern": "-", "replacement": "_", "mode": "first"}) == "a_b-c"
    assert replace({"text": "a-b-c", "pattern": "-", "replacement": "_", "mode": None}) == "a_b_c"

    # Empty/missing pattern is a no-op.
    assert replace({"text": "a-b-c", "pattern": "", "replacement": "_", "mode": "all"}) == "a-b-c"
    assert replace({"text": "a-b-c", "pattern": None, "replacement": "_", "mode": "all"}) == "a-b-c"

    # Numeric best-effort support: replace first N occurrences.
    assert replace({"text": "a-b-c-d", "pattern": "-", "replacement": "_", "mode": "2"}) == "a_b_c-d"

